## $Id: HelpParser.pm 1120 2012-01-02 13:46:14Z zw $

package Dursamvad::HelpParser;
use base Dursamvad::Parser;
use utf8;
use strict;

# Constructor needs a reference to a text widget and an arrayref with allowed tags
sub new {
my ($this, $tw, $at) = @_;
my $class = ref $this || $this;
my $self = new Dursamvad::Parser;
$self->{-TextWidget} = $tw;
$self->{-AllowedTags} = $at;
$self->{-Tags} = [];
$self->{-LastSeen} = 0;
return bless $self, $class;
}

# Text callback, gets a string with expanded entities
sub text {
my ($self, $text) = @_;
# No text allowed outside any text
return unless scalar @{$self->{-Tags}};
my $is_pre = grep {$_ eq 'pre'} @{$self->{-Tags}};
my $ls = $self->{-LastSeen};
# Normalize the text
$text =~ s/\r+\n/\n/g;
$text =~ s/\r/\n/g;
if ($ls eq 'startelement' || ($ls eq 'text' && !$is_pre)) {# remove leading spaces
  if ($is_pre) {# only the whole empty leading line for <pre>
    $text =~ s/^\s*\n+//;
  } else {
    $text =~ s/^\s*//;
  }
}
if (!$is_pre) {# normalize spaces
  $text =~ s/\s+/ /g;
  $text =~ s/^(.*?)\s*$/$1/s;
}
if (length $text) {# text output
  $text = " $text" if $ls eq 'text';
  $self->{-TextWidget}->insert('end', $text, $self->{-Tags});
  $self->{-LastSeen} = 'text';
}}

# Comment callback, gets the comment with expanded entities, the comment may be split
sub comment {
my ($self, $comment) = @_;
}

# Processing instruction callback, gets the comment with expanded entities, the comment may be
# split
sub processing_instruction {
my ($self, $pi) = @_;
}

# Start element callback, gets the element name and the hash ref to attributes
sub startelement {
my ($self, $elemname, $atrib) = @_;
# Convert the header element name
$elemname =~ s/^h\d+$/header/;
# Useful variables
my $known = grep {$_ eq $elemname} @{$self->{-AllowedTags}};
my $tags = $self->{-Tags};
# Insert a space if we are in text and the element is known
$self->{-TextWidget}->insert('end', ' ', $tags) if $known && $self->{-LastSeen} eq 'text';
# @id is always inserted as a mark
if (exists $atrib->{'id'} && $atrib->{'id'} ne '') {
  $self->{-TextWidget}->markSet($atrib->{'id'}, 'end');
}
# Ignore unknown elements
return unless $known;
push @$tags, $elemname;
if ($elemname eq 'a' && exists $atrib->{'href'} && $atrib->{'href'} ne '') {# link
  push @$tags, 'href';
  $self->{-TextWidget}->insert('end', $atrib->{'href'}, $tags);
  pop @$tags; # remove href
}
$self->{-LastSeen} = 'startelement';
}

# End element callback, gets the element name, pops out the tag and all nested tags
sub endelement {
my ($self, $elemname) = @_;
# Convert the header element name
$elemname =~ s/^h\d+$/header/;
my $tags = $self->{-Tags};
if (scalar grep {$_ eq $elemname} @$tags) {
  # Endline for selected elements
  if ($elemname =~ /^p|header$/) {
    $self->{-TextWidget}->insert('end', "\n", $tags);
  }
  # Pop out
  my $tag;
  do {
    $tag = pop @$tags;
  } until $tag eq $elemname;
}
$self->{-LastSeen} = 'endelement';
}

1;

__END__

=head1 NAME

Dursamvad::HelpParser

=head1 SYNOPSIS

SAX-like parser usable for simple formatting of the help files.

=head1 DESCRIPTION

The object is a simple SAX-like parser usable for parsing and simple formatting of the help files
that should be displayed in the C<Tk::ROText> widget. It is derived from C<Dursamvad::Parser> and
overrides the callback functions.

=head2 new

This is the constructor. It requires a reference to the C<Tk::ROText> widget and a reference to an
array of allowed elements. These elements are names of the configured tags.

=head2 processing_instruction, comment

These functions do nothing.

=head2 text

The function handles the text output. The texts are ignored outside elements. The leading and
trailing spaces are removed, newlines and sequences of spaces are normalized to a single space.
Word wrapping is assumed to be performed by the widget. Only the I<pre> elements is handled
differently. Spaces and newlines are preserved with the only exception that the leading empty lines
are deleted.

=head2 startelement

The function is invoked on element start. The C<@id> attribute is always inserted as a I<mark>
without checking the element name. The element name is then pushed to the array of tags if it is
known. The I<a> element is further examined. Its I<href> attribute is inserted to the text using
the I<href> tag.

=head2 endelement

If a tag of the given name exists in the list of tags, it is popped out together with all nested
tags. There is no need to examine the list of allowed tags. If a not allowed tag is met, it
certainly does not exist in the list of active tags and thence the function does nothing.

=head1 LICENSE

The package was developed as a part of the DURSAMVAD program, see C<dursamvad.pl> for further info.

=cut


