## $Id: Parser.pm 1118 2012-01-01 22:11:58Z zw $

package Dursamvad::Parser;
use utf8;
use strict;

# Constructor
sub new {
my $this = shift;
my $class = ref $this || $this;
return bless {-attr => {}}, $class;
}

# Expand basic entities, expects a string, returns a string with expanded entities
sub expandEntities {
my ($self, $string) = @_;
$string =~ s/\&lt;/</g;
$string =~ s/\&gt;/>/g;
$string =~ s/\&apos;/'/g;
$string =~ s/\&quot;/"/g;
$string =~ s/\&amp;/\&/g; # must be the last!
return $string;
}

# Text callback, gets a string with expanded entities
sub text {
my ($self, $text) = @_;
print "Text: <![$text]!>\n";
}

# Comment callback, gets the comment with expanded entities, the comment may be split
sub comment {
my ($self, $comment) = @_;
print "Comment: <!--$comment-->\n";
}

# Processing instruction callback, gets the comment with expanded entities, the comment may be
# split
sub processing_instruction {
my ($self, $pi) = @_;
print "Processing instruction: <?$pi?>\n";
}

# Start element callback, gets the element name and the hash ref to attributes
sub startelement {
my ($self, $elemname, $atrib) = @_;
print "Element: <$elemname>\n";
foreach (sort keys %$atrib) {
  my $val = $atrib->{$_};
  print "    $_=\"$val\"\n";
}}

# End element callback, gets the element name
sub endelement {
my ($self, $elemname) = @_;
print "Element: </$elemname>\n";
}

# Empty element callback, gets the same arguments as startelement, usually just calls the two
# previous functions
sub emptyelement {
my ($self, $elemname, $atrib) = @_;
$self->startelement($elemname, $atrib);
$self->endelement($elemname);
}

# Parse the string or an array of strings (if the arg is an arrayref), the function calls callbacks
sub parse {
my ($self, $string) = @_;
if (ref $string eq 'ARRAY') {
  foreach (@$string) {
    $self->parse($_);
  }
}
return if ref $string; # array already handled, other refs ignored
my $status = exists $self->{-status} ? $self->{-status} : 0;
while (length $string) {
  !$status && do {# in text
    my $text;
    if ($string =~ /<(!--|\?|\/?\w[-\w]*(:\w[-\w]*)?)/s) {
      $text = $`;  $string = $';  $status = $1;
    } else {
      $text = $string;  $string = '';
    }
    $self->text($self->expandEntities($text)) if $text;
    next;
  };
  $status eq '?' && do {# in processing instruction, allows both SGML and XML pi's
    my $pi;
    if ($string =~ /\??>/) {
      $pi = $`;  $string = $';  $status = 0;
    } else {
      $pi = $string;  $string = '';
    }
    $self->processing_instruction($self->expandEntities($pi));
    next;
  };
  $status eq '!--' && do {# in comment
    my $comment;
    if ($string =~ /-->/) {
      $comment = $`;  $string = $';  $status = 0;
    } else {
      $comment = $string;  $string = '';
    }
    $self->comment($self->expandEntities($comment));
    next
  };
  $status =~ /^\// && do {# within end element tag
    if ($string =~ />/) {
      $string = $';
      $status = substr($status, 1);
      $self->endelement($status);
      $status = 0;
    } else {
      $string = '';
    }
    next;
  };
  $string = $self->attributes($string);
  if ($string =~ /^\/?>/) {
    $self->startelement($status, $self->{-attr});
    $self->{-attr} = {};
    $self->endelement($status) if substr($string, 0, 1) eq '/';
    $string = $';  $status = 0;
  }
}
$self->{-status} = $status;
}

sub attributes {
my ($self, $string) = @_;
my $attr = $self->{-attr}; # must always exist
my $name = exists $self->{-aname} ? $self->{-aname} : '';
my $val = exists $self->{-avalue} ? $self->{-avalue} : '';
my $delim = exists $self->{-adelim} ? $self->{-adelim} : '';
my $err = exists $self->{-attrerr} ? $self->{-attrerr} : 0;
# The first /?> terminates the attributes even within the delimiters
my $ret = '';
if ($string =~ /(.*?)(\/?>.*)/s) {
  $string = $1;  $ret = $2;
}
while (length $string) {
  last if $err; # the whole contents ignored if an error occured within attributes
  $delim && do {# in the middle of a delimited value
    if ($string =~ /$delim/) {
      $val .= $`;  $string = $';
      $attr->{$name} = $self->expandEntities($val);
      $name = $val = $delim = '';
    } else {
      $val .= $string;  $string = '';
    }
    next;
  };
  $name && do {# name found, search for the (optional) value
    if ($string =~ /^\s*=\s*(.*)$/s) {# equal sign found
      $string = $1;
      if ($string =~ /^(['"])(.*)$/s) {
        $delim = $1;  $string = $2;  $val = '';
      } else {
        if ($string =~ /^\s*(\S?)\s*/s) {
          $attr->{$name} = $self->expandEntities($1);
          $name = $val = $delim = '';
          $string = $';
        }#???
      }
    } else {# no equal sign, postpone attribute search
      $string = '';  last;
    }
    next;
  };
  # Skip white spaces at the beginning
  $string =~ s/^\s*(.*)/$1/s;
  last unless $string;
  if ($string =~ /^(\w[-\w]*(:\w[-\w]*)?)/) {
    if ($name) {# an attribute without a value is pending
      $attr->{$name} = '';
    }
    $name = $1;
    $val = $delim = '';
    $string = $';
  } elsif ($string =~ /\S/) {
    $err = 1;  last;
  }
}
if ($ret) {# end of attributes found
  if ($name) {# attribute still pending, probably wrongly delimited or unterminated value
    $attr->{$name} = $self->expandEntities($val);
    $name = $val = $delim = '';
    $err = 0; # clear the error status at the end of attributes
  }
}
# Store values within the object
$self->{-aname} = $name;
$self->{-avalue} = $val;
$self->{-adelim} = $delim;
$self->{-attrerr} = $err;
return $ret;
}

1;

__END__

=head1 NAME

Dursamvad::Parser

=head1 SYNOPSIS

SAX-like parser handling possibly buggy XML and SGML strings

=head1 DESCRIPTION

The object is a simple SAX-like parser for XML and SGML strings that must not crash on errors. It
must somehow recover. The SAX-like events are printed to the standard output. It is assumed that
the object will be subclassed and the callback functions overriden.

=head2 new

This is the constructor. It does not expect any argument.

=head2 parse

This functionexpects a string (may contain several lines) or an arrayref to strings. The strings
are parsed. The function does not return anything but invoke the callback functions. You can call
the parse function several times. The object remembers its states so that the chunks need not be
well formed. It is even possible to break chunks in the middle of a multiline attribute value.

=head2 Callback functions

The callback functions get the information of the SAX events. They just output the events to the
standard output. They should be overridden in order to do enything useful.

=over 4

=item processing_instruction, comment, text

The function is invoked when a processing instruction, comment or text is found, respectivelly. The
function can be called several times if the contents is broken into chunks. This may be a problem
for processing instructions. The functions get the contents with expanded entities.

=item startelement

The function is invoked on element start. The function gets the element name and the hashref of
attributes. The function is invoked when the terminating > is found, thus the tag may be split to
several chunks.

=item endelement

The function is invoked on element end (when terminating > is found). It gets the element name.

=item emptyelement

The function is invoked when en empty element is found, ie the tag ending with />. The function
gets the sam arguments as C<startelement>. It calls C<startelement> and C<endelement> so that it
usually need not be overridden.

=back

=head1 LICENSE

The package was developed as a part of the DURSAMVAD program, see C<dursamvad.pl> for further info.

=cut


