## $Id: SerialPortInfo.pm 1151 2012-01-11 18:29:45Z zw $

package Dursamvad::SerialPortInfo;
use utf8;
use Device::SerialPort qw/:STAT/;
use Tk;
use Tk::ROText;

# The constructor takes max baudrate as a parameter
sub new {
my ($this, $MaxBaudrate, $MinBaudrate) = @_;
my $class = ref($this) || $this;
my $self = bless {-MaxBaudrate => $MaxBaudrate, -MinBaudrate => $MinBaudrate,
  -onSignal => {-bg => 'green', -relief => 'raised'},
  -offsignal => {-bg => 'red', -relief => 'sunken'},
  -sigmasks => {-DTR => TIOCM_DTR, -DSR => TIOCM_DSR, -DCD => TIOCM_CD,
                -RTS => TIOCM_RTS, -CTS => TIOCM_CTS, -RI => TIOCM_RI},
  -sigoptions => [qw/none DTR RTS/], -sigsend => [qw/DSR DCD CTS RI/],
  -varnames => [qw/baudrate databits stopbits parity handshake/]}, $class;
$self->noDevice;
return $self;
}

# Get the device name
sub getDeviceName {
return shift->{-name};
}

# Set device either from name or from another class instance
sub setDevice {
my ($self, $arg) = @_;
if (ref $arg) {
  $self->{-name} = $arg->{-name};
  $self->{-device} = $arg->{-device} if exists $arg->{-device};
} else {
  $self->{-name} = $arg;
  $self->{-device} = new Device::SerialPort($arg) or delete $self->{-device};
}
if (exists $self->{-device}) {
  my $port = $self->{-device};
  @{$self->{-baudrateOptions}} = sort {$b <=> $a}
  grep {$_ <= $self->{-MaxBaudrate} && $_ >= $self->{-MinBaudrate}} $port->baudrate;
  @{$self->{-databitsOptions}} = sort {$a <=> $b} $port->databits;
  @{$self->{-stopbitsOptions}} = sort {$a <=> $b} $port->stopbits;
  @{$self->{-parityOptions}} = $port->parity;
  @{$self->{-handshakeOptions}} = $port->handshake;
  $self->{-baudrate} = $port->baudrate;
  $self->{-databits} = $port->databits;
  $self->{-stopbits} = $port->stopbits;
  $self->{-parity} = $port->parity;
  $self->{-handshake} = $port->handshake;
  $self->dlgOptions;
} else {
  $self->noDevice;
}}

# Get the current baudrate or undef if a port does not exist
sub getBaudrate {
my $port = shift->getport || return undef;
my $bd = $port->baudrate || return undef;
return $bd;
}

# Reset the device
sub noDevice {
my $self = shift;
delete $self->{-device};
if (exists $self->{-commWindow} && exists $self->{-commWindow}->{-log}) {
  delete $self->{-commWindow}->{-log};
}
delete $self->{-commWindow};
$self->{-name} = '?';
$self->{-DTRmask} = $self->{-RTSmask} = 0;
foreach (@{self->{-sigsend}}) {
  $self->{"-$_"} = '';
}
foreach (@{$self->{-varnames}}) {
  $self->{"-$_".'Options'} = [];
  $self->{"-$_"} = '';
}
$self->dlgOptions;
}

# Internal function: update dialog options
sub dlgOptions {
my $self = shift;
return unless exists $self->{-dlg};
my $dlg = $self->{-dlg};
return unless 'HASH' eq ref $dlg;
foreach (@{$self->{-varnames}}) {
  my $nm = "-$_";
  $dlg->{$nm}->options($self->{$nm . 'Options'});
}}

# Set the dialog, expects the Toplevel or Frame; the grid manager will be used;
# it is assumed that the device name selection has already been created
sub setupDialog {
my ($self, $dlg) = @_;
my %dlg = {};
foreach (@{$self->{-varnames}}) {
  my $nm = "-$_";
  $dlg->Label(-text => ucfirst $_, -anchor => 'e', -padx => 3)
      ->grid($dlg{$nm} = $dlg->Optionmenu(
                             -options => $self->{"-$_".'Options'},
                             -variable => \$self->{"-$_"}
), -sticky => 'we');
}
$self->{-dlg} = \%dlg;
}

# This is a continuation of the above dialog for the monitor mode
sub setupSignalDialog {
my ($self, $dlg) = @_;
foreach (@{$self->{-sigsend}}) {
  $dlg->Label(-text => "Send $_ to", -anchor => 'e', -padx => 3)
      ->grid($dlg->Optionmenu(
                  -options => $self->{-sigoptions},
                  -variable => \$self->{"-$_"}
  ), -sticky => 'we');
}}

# Set the communication window, expects Toplevel or Frame reference, Id, text width, text height,
# common button options (width must be specified), common grid options, clear and log callbacks,
# reference to a status text variable. It returns the status of the DTR/RTS buttons according to
# the handshake (if handshake is rts, the buttons must be disabled). The function returns C<undef>
# if creation fails for any reason.

sub commWindow {
my ($self, $dlg, $id, $txtw, $txth, $optBut, $optGrid, $clear, $log, $stxt) = @_;
return undef unless exists $self->{-device} && exists $optBut->{-width};
my %commWindow = (-id => $id, -dlg => $dlg, -logging => 'Start log', -logfn => '');
my $wd = $optBut->{-width};
my %st = (-width => $wd, -fg => 'blue', -bg => 'cyan',
          -relief => 'ridge', -borderwidth => 3, -padx => 3);
$dlg->Label(-text => $self->{-name}, %st)->grid(
$dlg->Label(-text => $self->{-baudrate} . ' bd', %st),
$dlg->Label(-text => $self->{-databits} . ' databits', %st),
$dlg->Label(-text => $self->{-stopbits} . ' stopbits', %st),
$dlg->Label(-text => $self->{-parity} . ' prty', %st),
$dlg->Label(-text => $self->{-handshake} . ' hshk', %st),
%$optGrid);
$commWindow{-DTR} = $dlg->Label(-text => 'DTR', -width => $wd)->grid(
$commWindow{-DSR} = $dlg->Label(-text => 'DSR', -width => $wd),
$commWindow{-DCD} = $dlg->Label(-text => 'DCD', -width => $wd),
$commWindow{-RTS} = $dlg->Label(-text => 'RTS', -width => $wd),
$commWindow{-CTS} = $dlg->Label(-text => 'CTS', -width => $wd),
$commWindow{-RI}  = $dlg->Label(-text => 'RI', -width => $wd),
%$optGrid);
my $Text = $commWindow{-text} = $dlg->Scrolled('ROText', -wrap => 'none',
-width => $txtw, -height => $txth, -scrollbars => 'osoe')->grid(
'-', '-', '-', '-', '-', %$optGrid);
my $blog;
my $bclear = $dlg->Button(-text => 'Clear', -command => [$clear, $id], %$optBut)->grid(
$blog = $dlg->Button(-textvariable => \$commWindow{-logging}, -command => [$log, $id], %$optBut),
$dlg->Label(-textvariable => \$commWindow{-logfn}, -anchor => 'w', %st),
'-', '-', '-', %$optGrid);
$self->{-commWindow} = \%commWindow;
$Text->bind('<Control-Key-a>', sub {$Text->selectAll;});
foreach ([$bclear, 'Clear the serial port input window above the button'],
         [$blog, 'Toggle logging the serial port input']) {
  my ($b, $t) = @$_;
  $b->bind('<Any-Enter>', sub {$$stxt = $t});
  $b->bind('<Any-Leave>', sub {$$stxt = ''});
}
return $self->{-handshake} eq 'rts' ? 'disabled' : 'normal';
}

# Close dialog
sub closeDialog {
my $self = shift;
delete $self->{-dlg} if exists $self->{-dlg};
}

# Clear the text window
sub clearWindow {
my $self = shift;
return unless exists $self->{-commWindow} && exists $self->{-commWindow}->{-text};
$self->{-commWindow}->{-text}->delete('1.0', 'end');
}

# Toggle logging
sub logging {
my $self = shift;
return unless exists $self->{-commWindow};
$com = $self->{-commWindow};
if (exists $com->{-log}) {
  delete $com->{-log};
  $com->{-logfn} = '';
  $com->{-logging} = 'Start log';
} else {
  my $fn = $com->{-dlg}->getSaveFile(-title => 'Log file name',
    -filetypes => [['Text files', ['.txt', '.text']], ['All files', '*']]);
  if (defined $fn) {
    open($com->{-log}, ">$fn") && do {
      $com->{-logfn} = $fn;  $com->{-logging} = 'Stop log';
    } or do {
      delete $com->{-log};
    };
  }
}}

# Port loop
sub portloop {
my ($self, $otherport) = @_;
my $port = $self->getport or return;
return unless exists $self->{-commWindow};
$com = $self->{-commWindow};
my $lines = $port->modemlines;
if (!exists $comm->{-lines} || $comm->{-lines} != $lines) {
  foreach (keys %{$self->{-sigmasks}}) {
    my $sig = $self->{-sigmasks}->{$_} & $lines ? $self->{-onSignal} : $self->{-offsignal};
    $com->{$_}->configure(%$sig);
  }
  $com->{-lines} = $lines;
  if ($otherport) {
    $otherport->dtr_active(($lines & $self->{-DTRmask}) != 0) if $self->{-DTRmask};
    $otherport->rts_active(($lines & $self->{-RTSmask}) != 0) if $self->{-RTSmask};
  }
}
my ($count, $chars) = $port->read(255);
return unless $count;
# utf8::decode($chars) if utf8::valid($chars); # unreliable and may cause data loss
if ($otherport) {
  $otherport->write($chars);
}
if (exists $com->{-log}) {
  my $log = $com->{-log};
  print $log $chars;
}
$chars =~ s/\r+\n/\n/g;
$chars =~ s/\r/\n/g;
$chars =~ s/[\x00-\x08\x0b-\x1f]/ /g;
$com->{-text}->insert('end', $chars);
$com->{-text}->yviewMoveto(1);
}

# Send the initialization commands, return 0 on success or an error string
sub initialize {
my $self = shift;
return 'No port selected' unless exists $self->{-device};
$self->{-DTRmask} = $self->{-RTSmask} = 0; # just to be sure...
foreach (@{$self->{-sigsend}}) {
  my $sg = "-$_";
  $self->{-DTRmask} |= $self->{-sigmasks}->{$sg} if $self->{$sg} eq 'DTR';
  $self->{-RTSmask} |= $self->{-sigmasks}->{$sg} if $self->{$sg} eq 'RTS';
}
my $port = $self->{-device};
$port->baudrate($self->{-baudrate}) or return "Cannot set baudrate";
$port->databits($self->{-databits}) or return 'Cannot set databits';
$port->stopbits($self->{-stopbits}) or return 'Cannot set stopbits';
$port->parity($self->{-parity}) or return 'Cannot set parity';
$port->handshake($self->{-handshake}) or return 'Cannot set handshake';
$port->write_settings or return 'Cannot write settings';
$port->rts_active(0);  $port->dtr_active(0);
$port->read_char_time(0);
$port->read_const_time(2);
return 0;
}

# Internal function, returns port or undef
sub getport {
my $self = shift;
return exists $self->{-device} ? $self->{-device} : undef;
}

# Set DTR
sub dtr_active {
my ($self, $arg) = @_;
my $port = $self->getport or return;
return if $self->{-handshake} eq 'rts';
$port->dtr_active($arg);
}

# Set RTS
sub rts_active {
my ($self, $arg) = @_;
my $port = $self->getport or return;
return if $self->{-handshake} eq 'rts';
$port->rts_active($arg);
}

# Get modem lines
sub modemlines {
my $port = shift->getport or return undef;
return $port->modemlines;
}

# Read 255 characters, returns ($count, $chars)
sub read {
my $port = shift->getport or return undef;
my @ret = $port->read(255);
return @ret;
}

# Write a string, returns number of characters written
sub write {
my ($self, $string) = @_;
my $port = $self->getport or return 0;
return $port->write($string);
}

# Get specified modem line bit, return port and the bit
sub getModemBit {
my ($self, $bit) = @_;
my $port = $self->getport or return (undef, undef);
$bit = $bit & $port->modemlines;
return ($port, $bit);
}

# Flip DTR
sub flip_dtr {
my ($port, $bit) = shift->getModemBit(TIOCM_DTR);
return unless $port;
$port->dtr_active(!$bit);
}

# Flip RTS
sub flip_rts {
my ($port, $bit) = shift->getModemBit(TIOCM_RTS);
return unless $port;
$port->rts_active(!$bit);
}

# Pulse DTR for specified number of miliseconds
sub pulse_dtr {
my ($self, $arg) = @_;
my ($port, $bit) = $self->getModemBit(TIOCM_DTR);
return unless $port;
if ($bit) {
  $port->pulse_dtr_off($arg);
} else {
  $port->pulse_dtr_on($arg);
}}

# Pulse RTS for specified number of miliseconds
sub pulse_rts {
my ($self, $arg) = @_;
my ($port, $bit) = $self->getModemBit(TIOCM_RTS);
return unless $port;
if ($bit) {
  $port->pulse_rts_off($arg);
} else {
  $port->pulse_rts_on($arg);
}}

1;

__END__

=head1 NAME

Dursamvad::SerialPortInfo

=head1 SYNOPSIS

This class provides methods and parameters related to the serial port as well as the dialog for
parameter setting. The class is prepared for use with C<dursamvad.pl> and will hardly be useful
I<as is> in other applications.

=head1 DESCRIPTION

The class contains two types of methods. The first group consists of methods serving to build
dialog windows. The second group provides methods of accessing the port. The functions of the
second group will do nothing if the port is not open. It is assumed that the instance will not be
used for communication if C<initialize> fails.

=head2 new

The constructor accepts two parameters: the maximum and minimum allowed baudrate. These values will
be stored in the instance and used later when setting the options.

=head2 getDeviceName

The function returns the name of the port. If the port is not assigned, C<?> is returned.

=head2 setDevice

The function sets the device name and fills the valid options that may be used in the dialog
window. The argument is either a string or another C<Dursamvad::SerialPortInfo> instance. In the
latter case the port is not creaed because it may be locked. Instead the reference to an existing
port object is copied from the source instance.

=head2 getBaudrate

The function returns the current baudrate of the initialized port or C<undef> if the port does not
exist.

=head2 noDevice

The function sets the name to C<?>, deletes the port reference and initializes the port options to
empty values.

=head2 setupDialog

The function sets the dialog widget. Its argument must be a reference to a Toplevel or a Frame. The
options are put in two columns.

=head2 setupSignalDialog

This function sets the second part of the port selection widget. It is used when setting the port
for the monitor mode. Function C<setupDialog> must be called first with the same Toplevel or Frame.

=head2 commWindow

The function sets the communication window consisting of the port parameters, modem lines status
and the read/only text area with the B<Clear> button and C<Log> button. It needs the following
arguments: Toplevel or Frame reference, Id, text width, text height, common button options (width
must be specified), common grid options, clear and logging callbacks. It returns the status of the
DTR/RTS buttons according to the handshake (if handshake is rts, the buttons must be disabled). The
function returns C<undef> if creation fails for any reason.

=head2 closeDialog

The function is invoked when a dialog was closed. It means that the instance should remove all
variables that make sense only when the dialog exists.

=head2 clearWindow

The function clears the text window. It should be called from a callback. Notice that it does not
affect logging.

=head2 logging

The function toggles logging. If the port input is not being logged, it presents a dialog for file
name selection, otherwise the file is closed. The function assumes that the C<commWindow> function
has been called previously for window creation. The stored reference is ised as the dialog's
parent. The button text is changed and the file name is stored in the label close to the button. If
the window reference is not found, the function does silently nothing.

=head2 portloop

This function should be called periodically but only if the C<commWindow> function has already been
called. The function reads the modem status lines and displays the labels accordingly, It also
displays the input in the corresponding window. Optionally the input is logged. In the monitor mode
an argument specifying the other port must be given. The changes of modem lines as well as all
input strings are sent tothe other port.
port is needed. In such a case the changes of the modem lines are send

=head2 initialize

The function initializes the port. It returne 0 on success or an error message.

=head2 getport

This is an internal function returning the port reference or C<undef> if the port has not be
assigned.

=head2 read

The function attmpts to read 255 characters and returns the list of C<($count, $chars)>.

=head2 write

The function writes the string supplied as an argument and returns the number of characters
written.

=head2 modemlines

The function returns the modem lines status.

=head2 dtr_active, rts_active

The functions set the modem signal according to the value supplied as an argument. They do nothing
if the handshake is set to I<rts>.

=head2 flip_dtr, flip_rts

The function flips the status of DTR or RTS, respectively.

=head2 pulse_dtr, pulse_rts

The functions generate a pulse on the respective modem line the duration of which (in miliseconds)
is given as an argument. The functions first examine the modem line in order to decide whether the
pulse polarity should be ON or OFF.

=head1 BUGS

The constructor writes an error to the standard error output if the port cannot be open.

=head1 LICENSE

The package was developed as a part of the DURSAMVAD program, see C<dursamvad.pl> for further info.

=cut


