#!/usr/bin/perl
our $FileId = '$Id: dursamvad.pl 1153 2012-01-11 20:02:55Z zw $';

use utf8;
use POSIX;
use Device::SerialPort;
use File::Spec;
use Tk;
use Tk::ROText;
use Tk::BrowseEntry;

# Get program name and path. The program can hardly work on a non-UNIX OS,
# thus we use a forward slash only.
our $pgmname = $0;
our $pgmpath;
if ($pgmname =~ /^(.*)\/[^\/]+$/) {
  $pgmpath = $1;
} else {
  $pgmpath = '.';
}
$pgmname = File::Spec->rel2abs(-l $0 ? readlink $0 : $0, -l $0 ? $pgmpath : '.');
if ($pgmname =~ /^(.*)\/[^\/]+$/) {
  $pgmpath = $1;
} else {
  $pgmpath = '.';
}

# If the following module exists, it means that the modules are not installed
# in the system directories and the path has to be inserted.
unshift @INC, $pgmpath if -f "$pgmpath/Dursamvad/SerialPortInfo.pm";

# Require our modules
require Dursamvad::SerialPortInfo;
require Dursamvad::Help;

use strict;

# Constants
our $Version = '1.0';
our $SIOdelay = 10;
our $MaxBaudrate = 150_000;
our $MinBaudrate = 300;
our $EntryHistorySize = 15;
our $SendMaskLimit = 16384;

# Options
our %ButtonOptions = (-width => 15, -borderwidth => 3);
our %Gridoptions = (-sticky => 'nsew', -padx => 3, -pady => 3);
our %IceBearSoft = (-text => '(c) http://icebearsoft.euweb.cz [GPL v.2+]', -relief => 'ridge',
                    -fg => 'blue', -borderwidth => 3);

# Function prototypes
sub getDevices();
sub openMainWindow();
sub showDialog($);
sub setPort($$$);
sub cleanup();
sub StatusText($$);

# Global variables
our ($dlg, $loopId, $SendButton);
our $SendFileMask = 1;
our $statustext;

# Global terminal variables
our ($pulsetime, $echowin, $sigmode, $append, $logbut, $logfn, $log, $sendcounter, $snd);
our %append = (none => '', CR => "\r", LF => "\n", CRLF => "\r\n");
our $nohist = 0;

# Send STDERR to /dev/null
close STDERR;
open STDERR, '>', '/dev/null';

# Create empty serial ports
our @Port = (
  new Dursamvad::SerialPortInfo($MaxBaudrate, $MinBaudrate),
  new Dursamvad::SerialPortInfo($MaxBaudrate, $MinBaudrate),
);

# Create the main window
our $mw = new MainWindow (-title => "DURSAMVAD v. $Version");
$mw->CmdLine;
our $Help = new Dursamvad::Help("DURSAMVAD v. $Version", $mw);
$mw->Label(-text => 'RS232 terminal and monitor', -relief => 'ridge',
           -bg => 'cyan', -fg => 'blue', -borderwidth => 3)->grid('-', %Gridoptions);
$mw->Button(-text => 'Terminal mode', -command => \&setupTerminal, %ButtonOptions)->grid(
$mw->Button(-text => 'Monitor mode', -command => \&setupMonitor, %ButtonOptions), %Gridoptions);
$mw->Button(-text => 'Exit', -command => sub {exit}, %ButtonOptions)->grid(
$mw->Button(-text => 'Help', -command => sub {$Help->help('Main')}, %ButtonOptions), %Gridoptions);
$mw->Label(%IceBearSoft)->grid('-', %Gridoptions);
{
  my ($columns, $rows) = $mw->gridSize;
  for (my $i = 0;  $i < $columns; $i++) {
    $mw->gridColumnconfigure($i, -weight => 1);
  }
  for (my $i = 1;  $i < $rows - 1; $i++) {
    $mw->gridRowconfigure($i, -weight => 1);
  }
}
$mw->resizable(0, 0);
$mw->bind('<Control-q>', sub {exit});
$mw->bind('<Control-w>', sub {exit});
$mw->bind('<F1>', sub {$Help->help('Main')});

MainLoop;

# Return the list of RS232 devices
sub getDevices() {
my @devices;
opendir(DEV, '/dev') and do {
  @devices = map {"/dev/$_"} sort grep {/^tty(S|USB)\d+$/} readdir DEV;
  closedir DEV;
};
return @devices;
}

# Setup terminal device setting
sub setupTerminal {
$mw->withdraw;
my $portvar;
my @devices = getDevices;
if ($#devices < 0) {
  $mw->messageBox(-message => 'No RS232 device found, connect the device and try again!');
  openMainWindow;
  return;
}
my $dlgTerminal = $mw->Toplevel(-title => 'DURSAMVAD: Select a port for the terminal mode');
$dlgTerminal->withdraw;
$dlgTerminal->Label(-text => 'Device', -padx => 3, -anchor => 'e')->grid(
$dlgTerminal->Optionmenu(
            -command => [\&setPort, \$portvar, 0, 1],
            -options => \@devices,
            -variable => \$portvar), -sticky => 'we');
$Port[0]->setupDialog($dlgTerminal);
$dlgTerminal->Button(-text => 'OK', -command => \&startTerminal, %ButtonOptions)->grid(
$dlgTerminal->Button(-text => 'Cancel', -command => \&openMainWindow, %ButtonOptions),
$dlgTerminal->Button(-text => 'Help', -command => sub {$Help->help('Terminal_setup')}, %ButtonOptions));
$dlgTerminal->resizable(0, 0);
$dlgTerminal->bind('<F1>', sub {$Help->help('Terminal_setup')});
showDialog $dlgTerminal;
}

# Setup monitor device setting
sub setupMonitor {
$mw->withdraw;
my ($portvar0, $portvar1);
my @devices = getDevices;
if ($#devices < 0) {
  $mw->messageBox(-message => 'No RS232 device found, connect the device and try again!');
  openMainWindow;
  return;
}
my $dlgMonitor = $mw->Toplevel(-title => 'DURSAMVAD: Select ports for the monitor mode');
$dlgMonitor->withdraw;
my $frm0 = $dlgMonitor->Frame;
my $frm1 = $dlgMonitor->Frame;
$frm0->Label(-text => 'Device', -padx => 3, -anchor => 'e')->grid(
$frm0->Optionmenu(
     -command => [\&setPort, \$portvar0, 0, 1],
     -options => \@devices,
     -variable => \$portvar0), -sticky => 'we');
$Port[0]->setupDialog($frm0);
$Port[0]->setupSignalDialog($frm0);
$frm1->Label(-text => 'Device', -padx => 3, -anchor => 'e')->grid(
$frm1->Optionmenu(
     -command => [\&setPort, \$portvar1, 1, 0],
     -options => \@devices,
     -variable => \$portvar1), -sticky => 'we');
$Port[1]->setupDialog($frm1);
$Port[1]->setupSignalDialog($frm1);
my $bf = $dlgMonitor->Frame;
my %GB = (-side => 'left', -fill => 'x', -expand => 1);
$bf->Button(-text => 'OK',  -command => \&startMonitor, %ButtonOptions)->pack(%GB);
$bf->Button(-text => 'Cancel', -command => \&openMainWindow, %ButtonOptions)->pack(%GB);
$bf->Button(-text => 'Help',  -command => sub {$Help->help('Monitor_setup')}, %ButtonOptions)->pack(%GB);
$bf->pack(-side => 'bottom', -fill => 'x', -expand => 1);
$frm0->pack(-side => 'left', -fill => 'both', -expand => 1);
$frm1->pack(-side => 'left', -fill => 'both', -expand => 1);
$dlgMonitor->resizable(0, 0);
$dlgMonitor->bind('<F1>', sub {$Help->help('Monitor_setup')});
showDialog $dlgMonitor;
}

# Set the port device and options
sub setPort($$$) {
my ($nameref, $portx, $srcportx) = @_;
my $name = $$nameref;
my $port = $Port[$portx];
my $srcport = $Port[$srcportx];
if ($name ne $port->getDeviceName) {
  $dlg->withdraw if $dlg;
  $port->setDevice($name eq $srcport->getDeviceName ? $srcport : $name);
  if ($dlg) {
    $dlg->deiconify; $dlg->raise;
  }
}}

# Show the dialog window
sub showDialog($) {
$dlg = shift;
$dlg->protocol('WM_DELETE_WINDOW', \&Exit);
$dlg->deiconify;
$dlg->raise;
}

# Cleanup
sub cleanup() {
$snd = undef;
$loopId->cancel if $loopId;
$loopId = undef;
if ($dlg) {
  $dlg->destroy;
  $dlg = undef;
}
foreach (@Port) {
  $_->closeDialog;
  $_->noDevice;
}
$log = undef;
}

# Safe exit
sub Exit {
cleanup; exit;
}

# Reopen the main window
sub openMainWindow() {
cleanup;
$mw->deiconify;
$mw->raise;
}

# Start the Monitor
sub startMonitor {
my @name;
for (my $i = 0;  $i < 2;  $i++) {
  $name[$i] = $Port[$i]->getDeviceName;
}
if ($name[0] eq $name[1]) {
  $dlg->messageBox(-message => 'The ports must be different!', -icon => 'error');
  return;
}
for (my $i = 0;  $i < 2;  $i++) {
  my $result = $Port[$i]->initialize;
  if ($result) {
    $dlg->messageBox(-message => "($i) $name[$i]: $result", -icon => 'error');
    return;
  }
}
$dlg->withdraw;
my $mon = $mw->Toplevel(-title => 'DURSAMVAD: Monitor mode');
$mon->withdraw;
for (my $i = 0;  $i < 2;  $i++) {
  my $bs = $Port[$i]->commWindow($mon, $i, 90, 13, \%ButtonOptions, \%Gridoptions, \&clearWindow,
                                \&logging, \$statustext);
  if (!$bs) {
    $mon->destroy;
    $mw->messageBox(-message => "Cannot build the Monitor window for port #$i", -icon => 'error');
    openMainWindow();
    return;
  }
}
my ($bclose, $bexit, $bhelp);
$mon->Label(%IceBearSoft)->grid('-', '-',
$bclose = $mon->Button(-text => 'Close', -command => \&openMainWindow, %ButtonOptions),
$bexit = $mon->Button(-text => 'Exit', -command => \&Exit, %ButtonOptions),
$bhelp = $mon->Button(-text => 'Help', -command => sub {$Help->help('Monitor')}, %ButtonOptions),
%Gridoptions);
StatusText($mon, [[$bclose, 'Close the monitor window and return to the main window'],
                  [$bexit, 'Exit the program'],
                  [$bhelp, 'Display help']]);
my ($columns, $rows) = $mon->gridSize;
for (my $i = 0; $i < $columns; $i++) {
  $mon->gridColumnconfigure($i, -weight => 1);
}
foreach ((2, 6)) {
  $mon->gridRowconfigure($_, -weight => 1);
}
$mon->bind('<Control-q>', \&Exit);
$mon->bind('<Control-w>', \&openMainWindow);
$mon->bind('<F1>', sub {$Help->help('Monitor')});
showDialog $mon;
$loopId = $mon->repeat($SIOdelay, sub {
          $Port[0]->portloop($Port[1]);
          $Port[1]->portloop($Port[0]);});
}

# For BrowseEntry
our $string = '';
our $currententry = '';
our $beindex = 0;
our $be;

# Start the terminal
sub startTerminal {
my $port = $Port[0]; # only one port used in the terminal mode
my $result = $port->initialize;
if ($result) {# an error occured
  $dlg->messageBox(-message => $result, -icon => 'error');
  return;
}
$dlg->withdraw;
$SendFileMask = 2;
my $bd = $port->getBaudrate;
while ($bd <= $SendMaskLimit && $SendFileMask < 256) {
  $bd <<= 1;  $SendFileMask <<= 1;
}
$SendFileMask--;
my $term = $mw->Toplevel(-title => 'DURSAMVAD: Terminal mode');
$term->withdraw;
my $ButtonStatus = $port->commWindow($term, 0, 90, 15, \%ButtonOptions, \%Gridoptions,
  \&clearWindow, \&logging, \$statustext);
if (!$ButtonStatus) {
  $term->destroy;
  $mw->messageBox(-message => 'Cannot build the Terminal window', -icon => 'error');
  openMainWindow();
  return
}
$SendButton = 'Send file';
$pulsetime = 2000;
$log = undef;
$logbut = 'Start log';
$logfn = '';
$append = 'LF';
my ($bclear, $blog, $bmode, $ptime, $bdtr, $brts, $eol, $bsend, $bfile, $bclose, $bexit, $bhelp);
$echowin = $term->Scrolled('ROText', -wrap => 'none',
-width => 90, -height => 10, -scrollbars => 'osoe')->grid(
'-', '-', '-', '-', '-', %Gridoptions);
$bclear = $term->Button(-text => 'Clear', -command => sub {$echowin->delete('1.0', 'end');},
              %ButtonOptions)->grid(
$blog = $term->Button(-textvariable => \$logbut, -command => \&mainlog, %ButtonOptions),
$term->Label(-textvariable => \$logfn, -fg => 'blue', -bg => 'cyan', -padx => 3, -anchor => 'w',
             -relief => 'ridge', -borderwidth => 3),
'-', '-', '-', %Gridoptions);
$term->Label(-text => 'Signal mode', -padx => 3, -anchor => 'e')->grid(
$bmode = $term->Optionmenu(-variable => \$sigmode, -options => ['on/off', 'pulse']),
$term->Label(-text => 'Pulse time', -padx => 3, -anchor => 'e'),
$ptime = $term->Entry(-textvariable => \$pulsetime, -validate => 'key', -invalidcommand => sub {$mw->bell},
             -validatecommand => sub {$_[0] =~ /^\d*$/}),
$bdtr = $term->Button(-text => 'DTR', -state => $ButtonStatus, -command => sub {
  if ($sigmode eq 'pulse') {$port->pulse_dtr($pulsetime);} else {$port->flip_dtr;}}),
$brts = $term->Button(-text => 'RTS', -state => $ButtonStatus, -command => sub {
  if ($sigmode eq 'pulse') {$port->pulse_rts($pulsetime);} else {$port->flip_rts;}}),
%Gridoptions);
$be = $term->BrowseEntry(
    -variable => \$string,
    -browsecmd => \&browse,
    -autolimitheight => 1,
    -width => 50)->grid('-', '-', '-',
$eol = $term->Optionmenu(-variable => \$append, -options => [qw/none CR LF CRLF/]),
$bsend = $term->Button(-text => 'Send', -command => \&sendit, %ButtonOptions),
%Gridoptions);
$bfile = $term->Button(-textvariable => \$SendButton, -command => \&sendfile, %ButtonOptions)->grid(
$term->Label(%IceBearSoft), '-',
$bclose = $term->Button(-text => 'Close', -command => \&openMainWindow, %ButtonOptions),
$bexit = $term->Button(-text => 'Exit', -command => \&Exit, %ButtonOptions),
$bhelp = $term->Button(-text => 'Help', -command => sub {$Help->help('Terminal')}, %ButtonOptions),
%Gridoptions);
StatusText($term, [[$blog, 'Toggle logging of the user input'],
                   [$bclear, 'Clear the user input window'],
                   [$bmode, 'Switch between th on/off and pulse mode of the DTR and RTS buttons'],
                   [$ptime, 'Set the DTR/RTS pulse time in miliseconds'],
                   [$bdtr, 'Toggle or pulse the DTR line'],
                   [$brts, 'Toggle or pulse the RTS line'],
                   [$be, 'Enter the string to be sent'],
                   [$eol, 'Select what should be appended to the input string'],
                   [$bsend, 'Send the string with the appended value'],
                   [$bfile, 'Send a file or abort file send'],
                   [$bclose, 'Close the monitor window and return to the main window'],
                   [$bexit, 'Exit the program'],
                   [$bhelp, 'Display help']]);
my ($columns, $rows) = $term->gridSize;
for (my $i = 0; $i < $columns; $i++) {
  $term->gridColumnconfigure($i, -weight => 1);
}
foreach ((2, 4)) {
  $term->gridRowconfigure($_, -weight => 1);
}
$term->bind('<Control-Key-u>', sub {$string = 'http://icebearsoft.euweb.cz/dursamvad/';});
$term->bind('<Control-Key-g>', sub {
  $nohist = 1;
  $string = POSIX::strftime('%FT%TZ', gmtime);
  &sendit();
});
$term->bind('<Control-Key-t>', sub {
  $nohist = 1;
  $string = POSIX::strftime('%FT%T%z', localtime);
  &sendit();
});
$term->bind('<Control-q>', \&Exit);
$term->bind('<Control-w>', \&openMainWindow);
$term->bind('<F1>', sub {$Help->help('Terminal')});
$echowin->bind('<Control-Key-a>', sub {$echowin->selectAll;});
$be->bind('<Return>', \&sendit);
$be->focus;
showDialog $term;
$loopId = $dlg->repeat($SIOdelay, \&TerminalLoop);
}

# Build the status text field
sub StatusText($$) {
my ($dlg, $vars) = @_;
$statustext = '';
$dlg->Label(-textvariable => \$statustext, -fg => 'brown', -anchor => 'w',
            -padx => 6, -relief => 'sunken')
    ->grid('-', '-', '-', '-', '-', %Gridoptions);
foreach (@$vars) {
  my ($b, $t) = @$_;
  $b->bind('<Any-Enter>', sub {$statustext = $t});
  $b->bind('<Any-Leave>', sub {$statustext = ''});
}}

# Terminal loop
sub TerminalLoop {
my $port = $Port[0];
$port->portloop;
if ($snd) {
  if (!$sendcounter) {
    my $string = <$snd>;
    if ($string) {
      $port->write($string);
    } else {
      undef $snd;
      $SendButton = 'Send file';
    }
  }
  $sendcounter++; $sendcounter &= $SendFileMask;
}}

# Handle BrowseEntry
sub browse {
my @sel = $be->Subwidget('slistbox')->curselection;
if ($#sel < 0) {
  $currententry = '';
} else {
  $beindex = $sel[0];
  $currententry = $be->get($beindex);
}}

# Send and log the string
sub sendit {
if ($string && !$nohist) {
  $be->delete($beindex) if $string eq $currententry;
  $be->delete('end') if $be->Subwidget('slistbox')->size > $EntryHistorySize;
  $be->insert(0, $string);
}
$nohist = 0;
my $stg = $string . $append{$append};
$string = '';
$echowin->insert('end', $stg);
$echowin->yviewMoveto(1);
print $log $stg if $log;
$Port[0]->write($stg);
}

# Clear window
sub clearWindow {
my $id = shift;
$Port[$id]->clearWindow;
}

# Toggle logging
sub logging {
my $id = shift;
$Port[$id]->logging;
}

# Toggle logging of the user input
sub mainlog {
if ($log) {
  $log = undef;
  $logfn = '';
  $logbut = 'Start log';
} else {
  my $fn = $dlg->getSaveFile(-title => 'Log file name',
    -filetypes => [['Text files', ['.txt', '.text']], ['All files', '*']]);
  if (defined $fn) {
    open($log, ">$fn") && do {
      $logfn = $fn;  $logbut = 'Stop log';
    } or do {
      $log = undef;
    };
  }
}}

# Send a file, it is not logged, not displayed; abort if sending
sub sendfile {
if ($snd) {
  $snd = undef;
  $SendButton = 'Send file';
  return;
}
my $fn = $dlg->getOpenFile(-title => 'Select a file to send',
    -filetypes => [['Text files', ['.txt', '.text']], ['All files', '*']]);
$sendcounter = 0;
if (defined $fn) {
  open ($snd, $fn) or $dlg->messageBox(-message => "Cannot open $fn.", -icon => 'error');
  $SendButton = 'Abort send' if $snd;
}}

__END__

=head1 NAME

DURSAMVAD = dursamvad.pl

The name is derived from two Hindi words: I<dūr> meaning far or distant (pronounce with long I<u>
so that it rhymes with I<poor>) and I<samvād>, meaning discussion (the first I<a> should be
pronounced as I<u> in I<gun>, the second I<a> should be pronounced long as in I<grass>). The
transliteration in the Velthuis system or ITRANS would be dUrasaMvAda. Due to perldoc limitation
the Devanagari form is written on the web page only, see the link at the end of this manual.

=head1 SYNOPSIS

The program serves as a simple testing tool for RS232 communication. It can display not only the
received bytes but can also show the status of the modem lines as well as monitor communication of
two devices. It is possible to use USB to RS232 adapters.

The program was written and tested on Linux and may hopefully work on other Unix systems. It will
not work on other operating systems.

The program requires perl-Tk and Device::SerialPort. Both modules can be installed from CPAN. As
the first option you should use the packages included in your Linux distribution, try CPAN only if
the packages are not available. Specifically, perl-Tk on CPAN often requires newer perl than that
distributed with server-type Linux distros as RHEL or CentOS. Perl-Tk from CPAN may not work
properly in these distros while the rpm module, that can be installed and/or updated by yum, works
fine.

Specifically, in RHEL, CentOS, Fedora, you have to run

	yum install perl-Tk
	yum install perl-Device-SerialPort

In Debian, Ubuntu, Kubuntu, you have to run

	apt-get install perl-tk
	apt-get install libdevice-serialport-perl

=head1 DESCRIPTION

Some measurement devices still communicate using the RS232 interface. A few years ago RS232 port
appeared on all computers. Recently USB ports are preferred.

Measurement devices do not always contain communication software. It is useful to have possibility
to test communication manually using a terminal program. Programs as I<minicom> and I<picocom> are
helpful but do not display the modem lines. Program I<cutecom> does not allow to use USB (at least
the version distributed with Linux installed on my computer). I wrote this simple program in order
to solve these difficulties and in addition to learn how to develop a terminal with GUI.

The program offers two modes:

=over 4

=item Terminal mode

This mode allows you to communicate manually with an instrument (or with another computer).

=item Monitor mode

This mode allows monitoring of communication of two devices.

=back

All windows offer the following key bindings:

	Ctrl-q	Exit the program
	Ctrl-w	Close the window
	F1	Help

=head1 TERMINAL MODE

The terminal mode can be used for simple tests. Do not assume full terminal facilities.

=head2 Port selection for the terminal mode

When the terminal mode is chosen, a dialog with the port settings is displayed. First you have to
select the device. Other options are then offered according to the port capabilities. If other
options are not displayed, it means that the device name exists on the file system but the hardware
is not available. If your USB to RS232 adapter does not appear, you have probably inserted it too
late and the hotplug has not configured it yet. The remedy is easy, push the I<Cancel> button, wait
a while and then choose the terminal mode again.

=head2 Using the terminal mode

The upper row of the terminal window shows the current port name and options. The next row displays
the modem lines, the green background and raised relief means hat the line is on, the red
background and sunken relief means that the line is off.

The upper text area displays all incoming communication. The buttons just below allows you to clear
the window and to log incoming communication to a file. The file name vill be displayed below the
text window. There is no function allowing you to save the current contents of the window. However,
you can select any part of it and copy&paste it to another application. You can select the whole
contents by Ctrl-A or from a popup menu accessible by right-click.

The lower text area displays your input. Similarly you can clear it and log it to a file.

The DTR and RTS buttons allow you to switch the modem lines. Yo can choose between the on/off mode
and the pulse mode. Remember that the event loop is blocked during the pulse so that you will not
see any effect in the window. The buttons are disabled if hanshake is set to I<rts>.

The entry field below is used for your input. The history of the input is stored. The option
box to its right specifies what is appended to your output, the default is LF. The text is send
either by the Send button or by pressing I<Enter>. Three more keys are bound the the whole window
(not just to the entry field):

=over 4

=item *

Ctrl-g sends the current UTC date/time in the ISO 8601 format. The value is not stored in the
history.

=item *

Ctrl-t sends the curremt local data/time including the numeric time zone offset in the ISO 8601
format. As above tha string is not stored in the history.

=item *

Ctrl-u fills the entry field with the URL of this program, the previous contents is deleted. The
text is not sent, you can edit it before sending.

=back

The I<Send file> button allows you to send a text file. The contents will neither be shown in the
user input window nor will be logged. The event loop is not blocked during file send. It is
therefore possible to send text strings manually at the same time and they will be intermixed. No
order is guaranteed. Binary files or non-ASCII strings may even crash the whole program. While the
file is being sent, the very same button can be used to abort it. Abortion does not clear the
buffer, thus up to 4096 bytes may be sent after the transfer was aborted.

The I<Close> button closes the window and returns to the main window, the I<Exit> button terminates
the program.

=head1 MONITOR MODE

In order to use the monitor mode two interfaces are required. Remember that the program is used for
testing. You can easily intercept devices with a small amount of relatively slow communication. It
will be difficult or maybe even impossible to monitor high speed devices with a lot of sent
messages.

=head2 Port selection for the monitor mode

The dialog similarly offers selection of the ports and their options. It is basically the same as
for the terminal mode but you have to select two different devices, otherwise you will see an error
message when you press OK.

The input modem lines may be sent to an output lines of the other port. The default is to not send.
If you select to send several input lines to the same output line, logical OR will be used. A port
will not accept line settings from the other port if its handshake is set to I<rts>.

It is possible to use different settings on each interfaces. Thus the probram can serve as a rather
expensive connection between two incompatible devices.

=head2 Using the monitor mode

The monitor mode presents two areas each of which represents the modem lines and input texts from
the respective ports.

The monitor mode does not allow any user interaction. You can only clear any window or log the
input from any interface.

The I<Close> button closes the window and returns to the main window, the I<Exit> button terminates
the program.

=head1 INSTALLATION

The files can be placed into any directory provided the C<Dursamvad> directory with the
C<SerialPortInfo.pm> file is in the same directory as C<dursamvad.pl>. You can make a symbolic link
to C<dursamvad.pl> in a directory listed in the C<PATH> variable. A hard link will not work.
Alternativelly you can move C<dursamvad.pl> to a directory listed in C<PATH> and the C<Dursamvad>
directory to a place where perl looks for modules.

=head1 PIN LAYOUT

The following table shows the pin layout for Canon DB9 and DB25. The direction is writen from the
point of a computer.

	DB9  DB25  direction	abbrev	meaning
	 5     7		GND	Ground
	 2     3	in	RxD	Receive
	 3     2	out	TxD	Transmit
	 4    20	out	DTR	Data Terminal Ready
	 6     6	in	DSR	Data Set Ready
	 1     8	in	DCD	Data Carried Detect (sometimes CD = Carried Detect)
	 7     4	out	RTS	Request to Send
	 8     5	in	CTS	Clear to Send
	 9    22	in	RI	Ring Indicator

RS232 was used for intercommunication of two computers via a null modem where the signal grounds
were connected to each other and RxD of one connector was connected to TxD of the other. The modem
lines were rarely used. Two kinds of cables exist. In one kind DTR is connected to both DSR and DCD
and RTS is connected to CTS, in the other kind DTR is connected to both DSR and CTS and RTS is
connected to DCD. RI is either unused or connected to each other. The lines may be connected within
the connector or the lines from one conector are connected to the corresponding lines of the other
connector It is also possible to make a loopback connector where TxD is connected to RxD. This may
be useful for testing the hardware.

=head1 PROBLEMS

=over 4

=item *

The serial port works with bytes, not with characters. Only ASCII characters will be handled
correctly, accented and non-latin characters will turn into garbage.

=item *

Perl-Tk does not make use of any UNICODE rendering library, thus complex scripts such as Arabic,
Indic scripts, Chinese etc. are not displayed properly at all.

=item *

The program is intended for text communication. You can send short binary file in the terminal mode
(a long file may overful the buffer and bytes may be lost). It will be sent properly but received
binary data will not be corectly displayed. They may be correctly logged to a file but it is not
guaranteed. In some cases they may cause program crash.

=item *

The DTR/RTS pulses block the event loop.

=item *

Doubleclick does not select a file/directory in the open/save dialog, it is necessary to click the
Open/Save button. This is a limitation of the Tk widget.

=item *

The detailed manual is not yet written.

=back

=head1 TESTS

The program was verified by running a few tests. The following computers were used (the
abbreviations will be used later in the text):

=over 4

=item R1U

Desktop computer with RS232 and several USB interfaces running CentOS 4.7.

=item NB

ASUS Notebook with USB interfaces only running Fedora 13. The system reports existence of
/dev/ttyS1 (but not /dev/ttyS0). However, there is no slot provided.

=back

Connection to USB ports were done using AXAGO ADS-50 adapter with Prolific PL2303 chip. This
adapter is supported in Linux kernel most probably in all nowadays distros.

Both kinds of cables were used in tests, ie DTR->DSR+DCD and DTR->DSR+CTS (see above). A home made
loopback connector (connecting DTR+DSR+DCD and RTS+CTS) was also used. The loopback connector
allowed to connect optionally RI to RTS+CTS so that it was possible to test that the RI line is
properly read.

Communication with an old serial mouse was also tested.

=head2 Tests of the terminal mode

The first tests were performed by connecting the loopback connector both to the RS232 and to USB
ports via the AXAGO adapter. All commands were tested, ie sending a string, sending a file,
switching modem lines. As already written the DTR/RTS pulse locks the event loop, thus the display
is not chanched during the pulse and no change is visible.

Afterwards two ports of a single computer were connected, either RS232-USB os USB-USB (using two
AXAGO adapters) on R1U and two instances of the program were run. All tests as above were
performed. Simultaneous file sending in both directions at the same file was also tested. Since the
DTR/RTS pulse blocks sender's event loop only, the result of the pulse was visible on the
receiver's instance of the program.

A serial mouse was connected to a computer both directly to the RS232 port (R1U only) or via an
AXAGO adapter. It is known that the mouse works at 1200 bd, 7 bits, 1 stopbit, no parity, no
handshake. DTR and RTS are used to provide power. It was verified that a mouse sends data to the
computer.

The program author does not know how the handshake is implemented in the kernel driver and the
hardware. If set to I<none> or I<xon/xoff>, it works fine. If set to I<rts>, only USB with the AXAGO
adapter receives data both using a loopback adapter or using the DTR->DSR+DCD cable. If the
commercial DTR->DSR+CTS cable is used, communication with the I<rts> handshake is blocked.

=head2 Tests of the monitor mode

R1U and NB computers were connected either USB-RS232 os USB-USB. In case of USB-RS232 connection
the loopback connector was inserted via the second AXAGO adapter either to the NB computer or to
the R1U computer. In case of USB-USB connection the loopback connector was inserted to the RS232
port of R1U. The computer with the loopback connector was run in the monitor mode, the other
computer in the terminal mode. Data communication as well as transfer of modem line signals were
tested.

The loopback was then replaced with the serial mouse as mentioned in the terminal mode. Since the
monitor is passive, it is necessary to transfer the modem lines, eg DSR to DTR and CTS to RTS. This
way the remote terminal can set the DTR and RTS of the port where the serial mouse is connected. It
was verified that the mouse sends data to the monitor and the data are transfered to the terminal.

If a huge amount of data is being transfered, the text windows in the monitor mode need not be
refreshed. The program does its best not to lose data. The data to be displayed are sent to the
text widgets but sufficient time for refreshing the window may be unavailable. The widgets are
refreshed during a gap in communication. The monitor mode runs smoothly on a computer with two or
more CPU cores.

=head1 AUTHOR AND LICENSE

The author of the program is Zdeněk Wagner – Ice Bear Soft, the web page of the program is:

L<http://icebearsoft.euweb.cz/dursamvad/>

The program can be used, modified and redistributed under the GNU General Public License version 2
or newer.

=cut


